/*
* Copyright (c) 2024 Manuel Einfalt
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the “Software”), to
* deal in the Software without restriction, including without limitation the
* rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
* sell copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
* IN THE SOFTWARE. 
*/

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "wgif.h"

int gif_plainText(struct gif_file *gf, uint16_t left, uint16_t top, uint16_t width, uint16_t height, uint8_t cellWidth, uint8_t cellHeight, uint8_t fg, uint8_t bg, char *txt) {
    uint8_t buf[256];
    size_t block, sz, sent;

    buf[0] = 0x21;
    buf[1] = 1;
    buf[2] = 12;
    buf[3] = left & 0xff;
    buf[4] = (left >> 8) & 0xff;
    buf[5] = top & 0xff;
    buf[6] = (top >> 8) & 0xff;
    buf[7] = width & 0xff;
    buf[8] = (width >> 8) &0xff;
    buf[9] = height & 0xff;
    buf[10] = (height >> 8) & 0xff;
    buf[11] = cellWidth;
    buf[12] = cellHeight;
    buf[13] = fg;
    buf[14] = bg;
    gf->wr_fnct(buf, 15, gf->wparm);
    sent = 0;
    sz = strlen(txt);

    while (sent < sz) {
        if ((sz - sent) > 255) {
            block = 255;
        } else {
            block = sz - sent;
        }
        buf[0] = block;
        memcpy(buf + 1, txt + sent, block);
        gf->wr_fnct(buf, block + 1, gf->wparm);
        sent = sent + block;
    }
    buf[0] = 0;
    gf->wr_fnct(buf, 1, gf->wparm);

    return 1;
}

int gif_comment(struct gif_file *gf, char *str) {
    uint8_t buf[256];
    size_t sent = 0;
    size_t block;
    size_t blok0;
    size_t sz = strlen(str);

    buf[0] = 0x21;
    buf[1] = 0xfe;

    gf->wr_fnct(buf, 2, gf->wparm);

    while (sent < sz) {
        
        if ((sz - sent) > 255) {
            block = 255;
        } else {
            block = sz - sent;
        }

        buf[0] = block;
        memcpy(buf + 1, str + sent, block);
        gf->wr_fnct(buf, block + 1, gf->wparm);
        sent = sent + block;
    }
    buf[0] = 0;
    gf->wr_fnct(buf, 1, gf->wparm);

    return 1;
}

int gif_end(struct gif_file *gf) {
    int r;
    uint8_t e=';';

    r=gf->wr_fnct(&e, 1, gf->wparm);

    return r;
}

int gif_ext_loop(struct gif_file *gf, uint16_t loops) {
    uint8_t buf[19];

    buf[0]=0x21;
    buf[1]=0xff;
    buf[2]=0x0b;
    strcpy((char*)buf+3, "NETSCAPE2.0");
    buf[14]=3;
    buf[15]=1;
    buf[16]=loops&0xff;
    buf[17]=(loops>>8)&0xff;
    buf[18]=0x00;

    return gf->wr_fnct(buf, 19, gf->wparm);
}

int gif_graphicControlExtension(struct gif_file *gf, uint16_t delay_time, int transparent_color) {
    //TODO: disposal Method (1268)
    uint8_t p[8];

    p[0]='!';
    p[1]=0xf9;
    p[2]=4;
    p[3]=0;

    if (transparent_color!=-1) {
        p[3]=p[3]|1;
        p[6]=transparent_color;
    }

    p[4]=delay_time&0xff;
    p[5]=(delay_time>>8)&0xff;
    p[7]=0x00;

    return gf->wr_fnct(p, 8, gf->wparm);

}

int gif_init(struct gif_file *gf, uint16_t width, uint16_t height, uint32_t *colortab, int colortab_sz, uint8_t bgcol_idx) {
    int i, j;
    uint8_t p[6+7];
    int pot = 0;

    memcpy(p, "GIF89a", 6);
    p[6] = width & 0xff;
    p[7] = (width >> 8) & 0xff;
    p[8] = height & 0xff;
    p[9] = (height >> 8) & 0xff;
    p[10] = 0;

    gf->globColSz = 0; // TODO: oder hochgerundet?

    if (colortab_sz > 0) {
        do {
            if ((1 << (pot + 1)) >= colortab_sz) {
                break;
            }
            pot++;
        } while (1);

        gf->globColSz = 1 << (pot + 1);
        if (gf->globColSz < 4) {
            gf->globColSz = 4;
        }
        p[10] = 0xf0 | pot; //packed fields.
    }

    p[11] = bgcol_idx;
    p[12] = 0; // Aspect ratio.

    gf->wr_fnct(p, 6 + 7, gf->wparm);


    j = 1 << (pot + 1);

    if (colortab_sz) {
        for (i = 0; i < j; i++) {
            if (i >= colortab_sz) {
                p[0] = p[1] = p[2] = 0x00;
            }else {
                p[0] = (colortab[i] >> 16) & 0xff;
                p[1] = (colortab[i] >> 8) & 0xff;
                p[2] = colortab[i] & 0xff;
            }
            gf->wr_fnct(p, 3, gf->wparm);
        }
    }

    return 1;
}

int gif_image(struct gif_file *gf, uint16_t imgLeftPos, uint16_t imgTopPos, uint16_t width, uint16_t height, uint32_t *colorTab, int colorTabSz, uint8_t *pixmap) {
    int code_clear, code_stop;
    int next, next0;
    int x, y, pixel;
    uint8_t b[0x100+1];
    int bitc, bits;
    int t, bitlen;
    int c1[0xfff+1];
    unsigned c0[0xfff+1];
    //uint8_t *book[0xfff+1];
    //int len[0xfff+1];
    int i;
    int old;
    int have_code;
    int bit_count;
    int bytes_in;

    #define BITLEN(__X) do { \
        t=__X; \
        bitlen=0; \
        while (t) { \
            t=t>>1; \
            bitlen++; \
        } \
        if (bitlen>12) { \
            bitlen=12; \
        } \
    } while (0)

    #define BYTE2BUF(__X) do { \
        b[b[0]+1]=__X; \
        b[0]++; \
        if (b[0]==0xff) { \
            gf->wr_fnct(b, 0x100, gf->wparm); \
            b[0]=0; \
            if (bytes_in<0xff) { \
                BITLEN(next); \
                bits=bits|(code_clear<<bitc); \
                bitc=bitc+bitlen; \
                next=next0; \
            } \
            bytes_in=0; \
        } \
    } while (0)


    b[0]=0x2c;
    b[1]=imgLeftPos&0xff;
    b[2]=(imgLeftPos>>8)&0xff;
    b[3]=imgTopPos&0xff;
    b[4]=(imgTopPos>>8)&0xff;
    b[5]=width&0xff;
    b[6]=(width>>8)&0xff;
    b[7]=height&0xff;
    b[8]=(height>>8)&0xff;
    b[9]=0;

    bit_count=8; // weder globale noch lokale Farbtabelle == Systempalette (gif89a, "About Color Tables") (?)
    if (gf->globColSz>0) {
        t=gf->globColSz;
        bit_count=0;
        while (t) {
            t=t>>1;
            bit_count++;
        }
    } else if (colorTabSz>0) {
        i=0;
        while (colorTabSz>(1<<(i+1))) {
            i++;
        }
        bit_count=i+1;
        b[9]=0x80|i;
    }
    gf->wr_fnct(b, 10, gf->wparm);
    if (colorTabSz>0) {
        for (t=0;t<(1<<(i+1));t++) {
            if (t<colorTabSz) {
                b[0]=(colorTab[t]>>16)&0xff;
                b[1]=(colorTab[t]>>8)&0xff;
                b[2]=colorTab[t]&0xff;
            } else {
                b[0]=0;
                b[1]=0;
                b[2]=0;
            }
            gf->wr_fnct(b, 3, gf->wparm);
        }
    }
    if (bit_count<2) {
        bit_count=2;
    } else if (bit_count>8) {
        bit_count=8;
    }
    b[0]=bit_count; 
    gf->wr_fnct(b, 1, gf->wparm);

    code_clear=1<<bit_count;
    code_stop=code_clear+1;
    next=code_clear+2;

    bits=code_clear;
    bitc=bit_count+1;

    old=-1;
    bytes_in=0;

    for (i=0;i<=0xfff;i++) {
        c0[i]=i;
        c1[i]=0xffffffff;
    }

    next0=next;
    b[0]=0;

    for (y=0;y<height;y++) {
        for (x=0;x<width;x++) {
            pixel=pixmap[x+y*width];
            //printf("pixel(%i, %i)=%i\n", x, y, pixel);
            bytes_in++;
            if (old==-1) {
                old=pixel;
                continue;
            }
            have_code=0;
            for (i=0;i<next;i++) {
                if (c0[i]==old && c1[i]==pixel) {
                    //printf("\t@i=0x%x old=0x%x, pixel=%i\n", i, old, pixel);
                    /*printf("\tc0[i]==0x%x, c1[i]==0x%x. c0[old]==0x%x, c1[old]==0x%x\n",
                                    c0[i],       c1[i],        c0[old],       c1[old]);*/
                //if (len[i]==(len[old]+1) && memcmp(book[i], book[old], len[old])==0 && *(book[i]+len[old])==pixel) {
                    old=i;
                    have_code=1;
                    //printf("\told set to: %x\n", old);
                    break;
                }
            }
            if (have_code) {
                continue;
            }
            if (next>=0xfff) {
                bits=bits|(old<<bitc);
                bitc=bitc+12;
                old=pixel;
            } else {
                //len[next]=len[old]+1;
                //book[next]=malloc(len[next]);
                //memcpy(book[next], book[old], len[old]);
                //*(book[next]+len[old])=pixel;
                c0[next]=old;
                c1[next]=pixel;
                //printf("\tNEW(0x%x)=<0x%x, %x>\n",next, old, pixel);

                BITLEN(next-1);

                bits=bits|(old<<bitc);
                bitc=bitc+bitlen;
                next++;

                old=pixel;
            }
            
            while (bitc>=8) {
                t=bits&0xff;
                bits=bits>>8;
                bitc=bitc-8;

                BYTE2BUF(t);
            }
        }        
    }

    BITLEN(next-1);

    bits=bits|(old<<bitc);
    bitc=bitc+bitlen;
    
    BITLEN(next); // next statt next-1, weil nicht inkrementiert.
    bits=bits|(code_stop<<bitc);
    bitc=bitc+bitlen;

    while (bitc>=8) {
        t=bits&0xff;
        bits=bits>>8;
        bitc=bitc-8;
        BYTE2BUF(t);
    }


    if (bitc) {
        BYTE2BUF(bits);
    }
    if (b[0]!=0) {
        gf->wr_fnct(b, b[0]+1, gf->wparm);
    }
    b[0]=0;
    gf->wr_fnct(b, 1, gf->wparm);

    return 1;
}
